unit KnobParser;

{

   COPYRIGHT 1999 .. 2016 Blue Hell / Jan Punter

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License version 2 as
  published by the Free Software Foundation;

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

  For all listed email addresses :

    _dot. to be substituted by a dot      '.'
    2@t2  to be substituted by an at sign '@'


  Blue Hell is a trade mark owned by

    Jan Punter
    Oogstplein 6
    7545 HP Enschede
    the Netherlands
    http://www.iaf.nl/Users/BlueHell/
    http://bluehell.electro-music.com/
    j_dot.punter2@t2iaf_dot.nl

  All rights attributed to Blue Hell are owned by Jan Punter.
}


interface

uses

  System.Types, System.Classes, WinApi.Windows, System.SysUtils, System.Math, Vcl.Buttons, Vcl.Clipbrd, Vcl.Graphics,

  Globals, KnobsUtils, Knobs2013, FrmStore;


type

  TOnPatchReadError  = procedure( aSender: Tobject; aMsg: string; var Continue: Boolean) of object;
  TOnPatchWriteError = procedure( aSender: Tobject; aMsg: string; var Continue: Boolean) of object;


  TReaderState = record
    Pos    : Integer;
    LineNr : Integer;
    Offset : Integer;
  end;


  TPatchReader = class( TInterfacedObject, IKnobsPatchReader)
  strict private
  type
    TTokenType = (
      ttNothing,           // Token type not determined yet, start value
      ttError,             // Token did not start with a valid character
                           // [.()=;'] | (alpha) | (numeric) (any intro
                           // (white)* is always skipped).
      ttEof,               // Input exhausted without determining a token type.
      ttLParen,            // Left parenthesis  seen as the first non (white)
      ttRParen,            // Right parenthesis seen as the first non (white)
      ttEqual,             // Equial sign       seen as the first non (white)
      ttIdentifier,        // Identifier        seen : identifier = (alpha)(alpha-numeric)* .
      ttInteger,           // Integer           seen : integer = (numeric)*
      ttString             // String            seen : string = (quote)((any)*(quote)(quote)(any)*)*(quote) .
    );
  var
    FFileName             : string;
    FReadMode             : TKnobsReadMode;
    FPos                  : Integer;
    FLineNr               : Integer;
    FOffset               : Integer;
    FData                 : string;
    FWirePanel            : TKnobsWirePanel;
    FModule               : TKnobsCustomModule;
    FToken                : string;
    FTokenType            : TTokenType;
    FOnReadError          : TOnPatchReadError;
    FPatchVersion         : Integer;
    FWarnOnStructMismatch : Boolean;
  strict private
    procedure   SetWirePanel( const aWirePanel: TKnobsWirePanel);
  protected
    procedure   SaveState   ( var   aState: TReaderState);
    procedure   RestoreState( const aState: TReaderState);
    procedure   Clear;                                                                                          virtual;
    procedure   ReadError( const aMsg: string);                                                                 virtual;
    procedure   ReadErrorFmt( const aFmt: string; const anArgs: array of const);
    procedure   Warning( const aMsg: string);                                                                   virtual;
    procedure   WarningFmt( const aFmt: string; const anArgs: array of const);
    procedure   ReadToken;
    procedure   ExpectedSingle( aChar: Char);
    procedure   ExpectedIdentifier( const anIdent: string);
    procedure   ExpectedString( const aString: string);
    procedure   ExpectedInteger;
    procedure   ExpectedStringOrNumber;
    procedure   ExpectedIdentifierOrSingle( const anIdent: string; aSingle: Char);
    function    ReadLParen: Boolean;
    function    ReadRParen: Boolean;
    function    ReadEqual : Boolean;
    function    ReadIdentifier( const anIdent: string): Boolean;
    procedure   RequireLParen;
    procedure   RequireRParen;
    procedure   ScanRParen;
    procedure   RequireEqual;
    procedure   RequireIdentifier  ( const anIdent: string);
    function    ReadKVString       ( const anIdent: string; var aValue: string ): Boolean;
    function    ReadKVInteger      ( const anIdent: string; var aValue: Integer): Boolean;
    function    ReadKVIntegerOption( const anIdent: string; var aValue: Integer): Boolean;
    function    ReadClass        : string;
    function    ReadName         : string;
    function    ReadTitle        : string;
    function    ReadVersion      : Integer;
    function    ReadSrc          : string;
    function    ReadDst          : string;
    function    ReadType         : Integer;
    function    ReadLeft         : Integer;
    function    ReadTop          : Integer;
    function    ReadValue        ( var aValue : Integer): Boolean;
    function    ReadLockedOption ( var aValue : Boolean): Boolean;
    function    ReadCCOption     ( var aValue : Byte   ): Boolean;
    function    ReadStringValue  ( var aValue : string ): Boolean;
    function    ReadColorOption  ( var aValue : TColor ): Boolean;
    procedure   ReadWire;                                                                                       virtual;
    procedure   ReadWires;                                                                                      virtual;
    procedure   ReadDetail;                                                                                     virtual;
    procedure   ReadDetails;                                                                                    virtual;
    procedure   ReadModule;                                                                                     virtual;
    procedure   ReadModules;                                                                                    virtual;
    procedure   Read;                                                                                           virtual;
    function    TryRead: Boolean;                                                                               virtual;
  public
    constructor Create;                                                                                         virtual;
    destructor  Destroy;                                                                                       override;
    function    GetWarnOnStructMismatch: Boolean;
    procedure   SetWarnOnStructMismatch( aValue: Boolean);


    procedure   ReadParams(
      const aData      : string;
      const aWirePanel : TKnobsWirePanel;
      ReadStrict       : Boolean
    );                                                                                                          virtual;

    procedure   ReadString(
      const aData      : string;
      const aWirePanel : TKnobsWirePanel;
      aDistance        : TPoint;
      ReadStrict       : Boolean;
      MustDrag         : Boolean;
      CopyMidiCC       : Boolean;
      aReadMode        : TKnobsReadMode = rmReplace;
      const aFileName  : string         = ''
    );                                                                                                          virtual;

    procedure   ReadStrings(
      const aData      : TStrings;
      const aWirePanel : TKnobsWirePanel;
      ReadStrict       : Boolean;
      MustDrag         : Boolean;
      CopyMidiCC       : Boolean;
      aReadMode        : TKnobsReadMode = rmReplace;
      const aFileName  : string    = ''
    );                                                                                                          virtual;

    procedure   ReadFile(
      const aFileName  : string;
      const aWirePanel : TKnobsWirePanel;
      ReadStrict       : Boolean;
      MustDrag         : Boolean;
      CopyMidiCC       : Boolean;
      aReadMode        : TKnobsReadMode = rmReplace
    );                                                                                                          virtual;

    procedure   ImportFile(
      const aFileName  : string;
      const aWirePanel : TKnobsWirePanel;
      ReadStrict       : Boolean;
      MustDrag         : Boolean;
      CopyMidiCC       : Boolean;
      aReadMode        : TKnobsReadMode = rmReplace
    );                                                                                                          virtual;

    procedure   ReadFromClipboard(
      const aWirePanel : TKnobsWirePanel;
      ReadStrict       : Boolean;
      MustDrag         : Boolean;
      CopyMidiCC       : Boolean;
      aReadMode        : TKnobsReadMode = rmReplace
    );                                                                                                          virtual;

    procedure   ParamsFromFile(
      const aFileName  : string;
      const aWirePanel : TKnobsWirePanel;
      ReadStrict       : Boolean
    );                                                                                                          virtual;

    procedure   ParamsFromClipboard(
      const aWirePanel : TKnobsWirePanel;
      ReadStrict       : Boolean
    );                                                                                                          virtual;
  public
    property    FileName             : string  read FFileName;
    property    LineNr               : Integer read FLineNr;
    property    Offset               : Integer read FOffset;
    property    WarnOnStructMismatch : Boolean read GetWarnOnStructMismatch write SetWarnOnStructMismatch;
  public
    property    OnReadError : TOnPatchReadError read FOnReadError write FOnReadError;
  end;


  TPatchWriter = class( TInterfacedObject, IKnobsPatchWriter)
  strict private
  var
    FFileName     : string;
    FWriteMode    : TKnobsWriteMode;
    FWirePanel    : TKnobsWirePanel;
    FModule       : TKnobsCustomModule;
    FComponent    : TComponent;
    FWire         : TKnobsWire;
    FOnWriteError : TOnPatchWriteError;
    FResult       : string;
    FLineBreak    : string;
  protected
    procedure   Clear;                                                                                          virtual;
    procedure   Write          ( anInd: Integer; const S: string);
    procedure   WriteLn        ( anInd: Integer; const S: string);
    procedure   WriteWire      ( anInd: Integer);                                                               virtual;
    procedure   WriteWires     ( anInd: Integer);                                                               virtual;
    procedure   WriteDetailData( anInd: Integer; const aValue, aCC: Integer; aLocked: Boolean; const aSuffix: string); overload; virtual;
    procedure   WriteDetailData( anInd: Integer; const aValue: string );                                               overload; virtual;
    procedure   WriteDetail    ( anInd: Integer);                                                               virtual;
    procedure   WriteDetails   ( anInd: Integer);                                                               virtual;
    procedure   WriteModule    ( anInd, anXOffset, anYOffset: Integer);                                         virtual;
    procedure   WriteModules   ( anInd: Integer);                                                               virtual;
    procedure   WritePatch;                                                                                     virtual;
    function    TryWrite: Boolean;                                                                              virtual;
  public
    constructor Create;                                                                                         virtual;
    destructor  Destroy;                                                                                       override;

    function    WriteString(
      const aWirePanel: TKnobsWirePanel;
      const aFileName : string     = '';
      aWriteMode      : TKnobsWriteMode = wmAll
    ): string;                                                                                                  virtual;

    function    WriteStrings(
      const aWirePanel : TKnobsWirePanel;
      const aFileName  : string     = '';
      aWriteMode       : TKnobsWriteMode = wmAll
    ): TStrings;                                                                                                virtual;

    procedure   WriteFile(
      const aWirePanel : TKnobsWirePanel;
      const aFileName  : string;
      aWriteMode       : TKnobsWriteMode = wmAll
    );                                                                                                          virtual;

    procedure   WriteToClipboard(
      const aWirePanel : TKnobswirePanel;
      aWriteMode       : TKnobsWriteMode = wmAll
    );                                                                                                          virtual;

  public
    property    FileName  : string read FFileName;
    property    LineBreak : string read FLineBreak write FLineBreak;
  public
    property    OnWriteError : TOnPatchWriteError read FOnWriteError write FOnWriteError;
  end;



implementation


const

  STRUCT_MISMATCH_NO_PASTE = 'Structures differ, can not paste parameters';


  function QuoteEscape( const aValue: string): string;
  var
    i : Integer;
  begin
    Result := '';
    for i := 1 to Length( aValue)
    do begin
      if aValue[ i] = ''''
      then Result := Result + ''''''
      else Result := Result + aValue[ i];
    end;
  end;


  function UnQuoteEscape( const aValue: string): string;
  type
    TState = (
      stNormal,
      stInQuote
    );
  var
    State : TState;
    i     : Integer;
  begin
    State  := stNormal;
    Result := '';

    for i := 0 to Length( aValue)
    do begin
      case State of

        stNormal :
          begin
            if aValue[ i] = ''''
            then State := stInQuote
            else Result := Result + aValue[ i]
          end;

        stInQuote :
          begin
            if aValue[ i] = ''''
            then Result := Result + ''''
            else Result := Result + '''' + aValue[ i];

            State := stNormal;
          end;
      end;
    end;
  end;


{ ========
  TPatchReader = class( TInterfacedObject, IKnobsPatchReader)
  strict private
  type
    TTokenType = (
      ttNothing,           // Token type not determined yet, start value
      ttError,             // Token did not start with a valid character
                           // [.()=;'] | (alpha) | (numeric) (any intro
                           // (white)* is always skipped).
      ttEof,               // Input exhausted without determining a token type.
      ttLParen,            // Left parenthesis  seen as the first non (white)
      ttRParen,            // Right parenthesis seen as the first non (white)
      ttEqual,             // Equial sign       seen as the first non (white)
      ttIdentifier,        // Identifier        seen : identifier = (alpha)(alpha-numeric)* .
      ttInteger,           // Integer           seen : integer = (numeric)*
      ttString             // String            seen : string = (quote)((any)*(quote)(quote)(any)*)*(quote) .
    );
  var
    FFileName             : string;
    FReadMode             : TKnobsReadMode;
    FPos                  : Integer;
    FLineNr               : Integer;
    FOffset               : Integer;
    FData                 : string;
    FWirePanel            : TKnobsWirePanel;
    FModule               : TKnobsCustomModule;
    FToken                : string;
    FTokenType            : TTokenType;
    FOnReadError          : TOnPatchReadError;
    FPatchVersion         : Integer;
    FWarnOnStructMismatch : Boolean;
  public
    property    FileName             : string  read FFileName;
    property    LineNr               : Integer read FLineNr;
    property    Offset               : Integer read FOffset;
    property    WarnOnStructMismatch : Boolean read FWarnOnStructMismatch write FWarnOnStructMismatch;
  public
    property    OnReadError : TOnPatchReadError read FOnReadError write FOnReadError;
  strict private
}

    procedure   TPatchReader.SetWirePanel( const aWirePanel: TKnobsWirePanel);
    begin
      Clear;
      FWirePanel := aWirePanel;
    end;


//  protected

    procedure   TPatchReader.SaveState( var aState: TReaderState);
    begin
      aState.Pos    := FPos;
      aState.LineNr := LineNr;
      aState.Offset := Offset;
    end;


    procedure   TPatchReader.RestoreState( const aState: TReaderState);
    begin
      FPos    := aState.Pos;
      FLineNr := aState.LineNr;
      FOffset := aState.Offset;
    end;


    procedure   TPatchReader.Clear; // virtual;
    begin
      if Assigned( FWirePanel)
      then FreeAndNil( FWirePanel);
    end;


    procedure   TPatchReader.ReadError( const aMsg: string); // virtual;
    var
      S : string;
    begin
      if FileName = ''
      then
        S :=
          Format(
            'parse error:'^M^M'line = %d'^M'offset = %d'^M^M'%s',
            [
              LineNr,
              Offset,
              aMsg
            ]
          )
      else
        S :=
          Format(
            'parse error:'^M^M'file = ''%s'''^M'line = %d'^M'offset = %d'^M^M'%s',
            [
              FileName,
              LineNr,
              Offset,
              aMsg
            ]
          );

      raise EKnobsCorruptPatch.Create( S);
    end;


    procedure   TPatchReader.ReadErrorFmt( const aFmt: string; const anArgs: array of const);
    begin
      ReadError( Format( aFmt, anArgs));
    end;


    procedure   TPatchReader.Warning( const aMsg: string); // virtual;
    begin
      raise EKnobsWarning.CreateFmt( 'Warning: %s', [ aMsg]);
    end;


    procedure   TPatchReader.WarningFmt( const aFmt: string; const anArgs: array of const);
    begin
      Warning( Format( aFmt, anArgs));
    end;


    procedure   TPatchReader.ReadToken;
    const
      quote         = '''';                               // Single quotes used for strings, escaped as ''
      lparen        = '(';                                // Left  parenthesis
      rparen        = ')';                                // Right parenthesis
      equal         = '=';                                // Equals sign, assignment operator
      cr            = ^M;                                 // Carriage return, line feeds are ignored
      white         = [ #$00 .. #$20, #$7f .. #$ff];      // White space
      alpha         = [ 'A' .. 'Z', 'a' .. 'z'];          // Alpha characters
      numeric       = [ '0' .. '9', '-'];                 // Numeric characters
      alpha_numeric = alpha + numeric;                    // Alpha-numeric characters
    type
      TState = (
        stInitial,                                        // Initial parser state, not recognized anything yet
        stDone,                                           // Parsing completed, current token and token type are set
        stString,                                         // In string parsing, after seeing a string quote, escapes allowed
        stIdentifier,                                     // In identifier parsing, after seeing an alpha character
        stInteger                                         // In integer scanning, after seeing a numeric character
      );
    var
      i     : Integer;
      State : TState;
    begin
      Assert( FPos > 0, 'Can''t start token parsing upfront input :shock:');
      State      := stInitial;
      FTokenType := ttNothing;
      FToken     := '';
      i          := FPos;

      while ( i <= Length( FData)) and ( State <> stDone)
      do begin
        if FData[ i] = cr
        then begin
          inc( FLineNr);
          FOffset := 0;
        end
        else Inc( FOffset);

        case State of
          stInitial :
            begin
              case FData[ i] of
                lparen : begin State := stDone; FTokenType := ttLParen; FToken := '('; end;
                rparen : begin State := stDone; FTokenType := ttRParen; FToken := ')'; end;
                equal  : begin State := stDone; FTokenType := ttEqual;  FToken := '='; end;
                quote  : State := stString;
                else begin
                  if CharInSet( FData[ i], numeric)
                  then begin State := stInteger;    FToken := FData[ i]; end
                  else if CharInSet( Fdata[ i], alpha)
                  then begin State := stIdentifier; FToken := FData[ i]; end
                  else if not CharInSet( FData[ i], white)
                  then begin State := stDone; FToken := ''; FTokenType := ttError end;
                end;
              end;
            end;

          stDone :
            begin
              Assert( False, 'We should not be here ...');
            end;

          stString :
            begin
              if FData[ i] = quote
              then begin
                if ( i < Length( FData)) and ( FData[ i + 1] = quote)
                then begin
                  FToken := FToken + quote;
                  inc( i);
                end
                else begin
                  State      := stDone;
                  FTokenType := ttString;
                end;
              end
              else FToken := FToken + FData[ i];
            end;

          stIdentifier :
            begin
              if CharInSet( FData[ i], alpha_numeric)
              then FToken := FToken + FData[ i]
              else begin
                dec( i);
                State      := stDone;
                FTokenType := ttIdentifier;
              end;
            end;

          stInteger :
            begin
              if CharInSet( FData[ i], numeric)
              then FToken := FToken + FData[ i]
              else begin
                dec( i);
                State      := stDone;
                FTokenType := ttInteger;
              end;
            end;
        end;

        inc( i);
      end;

      if i > Length( FData)
      then FTokenType := ttEof;

      FPos := i;
    end;


//  protected

    procedure   TPatchReader.ExpectedSingle( aChar: Char);
    begin
      ReadErrorFmt( '''%s'' expected', [ aChar]);
    end;


    procedure   TPatchReader.ExpectedIdentifier( const anIdent: string);
    begin
      if anIdent = ''
      then ReadError   ( '<identifier> expected')
      else ReadErrorFmt( '''%s'' expected', [ anIdent]);
    end;


    procedure   TPatchReader.ExpectedString( const aString: string);
    begin
      if aString = ''
      then ReadError   ( '<string> expected')
      else ReadErrorFmt( '''%s'' expected', [ aString]);
    end;


    procedure   TPatchReader.ExpectedStringOrNumber;
    begin
      ReadError( '<string> or <integer> expected');
    end;


    procedure   TPatchReader.ExpectedIdentifierOrSingle( const anIdent: string; aSingle: Char);
    begin
      if anIdent = ''
      then ReadErrorFmt( '<identifier> or ''%s'' expected', [ aSingle])
      else ReadErrorFmt( '''%s'' or ''%s'' expected', [ anIdent, aSingle]);
    end;


    procedure   TPatchReader.ExpectedInteger;
    begin
      ReadError( 'integer value expected');
    end;


    function    TPatchReader.ReadLParen : Boolean;
    begin
      ReadToken;
      Result := FTokenType = ttLParen;
    end;


    function    TPatchReader.ReadRParen : Boolean;
    begin
      ReadToken;
      Result := FTokenType = ttRParen;
    end;


    function    TPatchReader.ReadEqual : Boolean;
    begin
      ReadToken;
      Result := FTokenType = ttEqual;
    end;


    function    TPatchReader.ReadIdentifier( const anIdent: string): Boolean;
    begin
      ReadToken;
      Result := ( FTokenType = ttIdentifier) and SameText( FToken, anIdent);
    end;


    procedure   TPatchReader.RequireLParen;
    begin
      if not ReadLParen
      then ExpectedSingle( '(');
    end;


    procedure   TPatchReader.RequireRParen;
    begin
      if not ReadRParen
      then ExpectedSingle( ')');
    end;


    procedure   TPatchReader.ScanRParen;
    begin
      ReadToken;

      while not ( FTokenType in [ ttRParen, ttEof, ttError])
      do ReadToken;

      if FTokenType <> ttRParen
      then ExpectedSingle( ')');
    end;


    procedure   TPatchReader.RequireEqual;
    begin
      if not ReadEqual
      then ExpectedSingle( '=');
    end;


    procedure   TPatchReader.RequireIdentifier( const anIdent: string);
    begin
      if not ReadIdentifier( anIdent)
      then ExpectedIdentifier( anIdent);
    end;


    function    TPatchReader.ReadKVString( const anIdent: string; var aValue: string): Boolean;
    // Scan for 'anIdent = <stringval>;'
    begin
      Result := False;
      RequireIdentifier( anIdent);
      RequireEqual;
      ReadToken;
      aValue := FToken;

      if FTokenType = ttString
      then Result := True;
    end;


    function    TPatchReader.ReadKVInteger( const anIdent: string; var aValue: Integer): Boolean;
    // Scan for 'anIdent = <intval>;'
    var
      aVal : string;
    begin
      Result := False;
      RequireIdentifier( anIdent);
      RequireEqual;
      ReadToken;
      aVal := FToken;

      if FTokenType = ttInteger
      then begin
        aValue := StrToIntDef( aVal, - MaxInt);
        Result := aValue <> - MaxInt;
      end
      else aValue := 0;
    end;


    function    TPatchReader.ReadKVIntegerOption( const anIdent: string; var aValue: Integer): Boolean;
    // Scan for 'anIdent = <intval>;'
    var
      aVal   : string;
      aState : TReaderState;
    begin
      Result := False;
      SaveState( aState);

      if ReadIdentifier( anIdent)
      then begin
        RequireEqual;
        ReadToken;
        aVal := FToken;

        if FTokenType = ttInteger
        then begin
          aValue := StrToIntDef( aVal, - MaxInt);
          Result := aValue <> - MaxInt;
        end
        else aValue := 0;
      end
      else begin
        RestoreState( aState);
        aValue := 0;
      end;
    end;


    function    TPatchReader.ReadClass : string;
    // Scan for 'class = <stringval>;'
    begin
      if not ReadKVString( 'class', Result)
      then ExpectedString( '');
    end;


    function    TPatchReader.ReadName : string;
    // Scan for 'name = <stringval>;'
    begin
      if not ReadKVString( 'name', Result)
      then ExpectedString( '');
    end;


    function    TPatchReader.ReadTitle : string;
    // Scan for 'title = <stringval>;'
    begin
      if not ReadKVString( 'title', Result)
      then ExpectedString( '');
    end;


    function    TPatchReader.ReadVersion: Integer;
    var
      aValue : string;
      aState : TReaderState;
    begin
      Result := 0;
      SaveState( aState);

      if ReadIdentifier( 'version')
      then begin
        RequireEqual;
        ReadToken;
        aValue := FToken;

        if FTokenType = ttInteger
        then Result := StrToInt( aValue)
        else begin
          RestoreState( aState);
          ExpectedInteger;
        end;
      end
      else RestoreState( aState);
    end;


    function    TPatchReader.ReadSrc : string;
    begin
      if not ReadKVString( 'src', Result)
      then ExpectedString( '');
    end;


    function    TPatchReader.ReadDst : string;
    begin
      if not ReadKVString( 'dst', Result)
      then ExpectedString( '');
    end;


    function    TPatchReader.ReadType : Integer;
    // Scan for 'type = <intval>;'
    begin
      if not ReadKVInteger( 'type', Result)
      then ExpectedInteger;
    end;


    function    TPatchReader.ReadLeft : Integer;
    // Scan for 'left = <intval>;'
    begin
      if not ReadKVInteger( 'left', Result)
      then ExpectedInteger;
    end;


    function    TPatchReader.ReadTop : Integer;
    // Scan for 'top = <intval>;'
    begin
      if not ReadKVInteger( 'top', Result)
      then ExpectedInteger;
    end;


    function    TPatchReader.ReadValue( var aValue : Integer): Boolean;
    // Scan for 'value = <Integer>;'
    begin
      Result := ReadKVInteger( 'value', aValue);
    end;


    function    TPatchReader.ReadLockedOption( var aValue: Boolean): Boolean;
    // Scan for 'locked = <integer>'
    var
      aState : TReaderState;
    begin
      Result := False;
      aValue := False;
      SaveState( aState);

      if ReadIdentifier( 'locked')
      then begin
        aValue := True;
        Result := True;
      end
      else RestoreState( aState);
    end;


    function    TPatchReader.ReadCCOption( var aValue : Byte): Boolean;
    // Scan for 'locked = <byte>'
    var
      aState : TReaderState;
      anInt  : Integer;
    begin
      Result := False;
      aValue := 0;
      SaveState( aState);

      if ReadKVIntegerOption( 'cc', anInt)
      then begin
        if ( anInt >= 0) and ( anInt < 256)
        then begin
          aValue := anInt;
          Result := True;
        end;
      end
      else RestoreState( aState);
    end;


    function    TPatchReader.ReadStringValue( var aValue : string): Boolean;
    // Scan for 'value = <string>;'
    begin
      Result := ReadKVString( 'value', aValue);
    end;

    function    TPatchReader.ReadColorOption( var aValue: TColor): Boolean;
    var
      aStrVal : string;
      aState  : TReaderState;
    begin
      Result := False;
      aValue := clYellow;
      SaveState( aState);

      if ReadIdentifier( 'color')
      then begin
        RequireEqual;
        ReadToken;
        aStrVal := FToken;

        if FTokenType = ttInteger
        then begin
          aValue := TColor( StrToInt( aStrVal));
          Result := True;
        end
        else begin
          RestoreState( aState);
          ExpectedInteger;
        end;
      end
      else RestoreState( aState);
    end;


    procedure   TPatchReader.ReadWire; // virtual;
    var
      aSrc : string;
      aDst : string;
    begin
      RequireIdentifier( 'wire');
      RequireLParen;
      ReadClass; // Result ignored
      aSrc := ReadSrc;
      aDst := ReadDst;
      RequireRParen;
      FWirePanel.AddWire( aSrc, aDst);
    end;


    procedure   TPatchReader.ReadWires; // virtual;
    var
      Done   : Boolean;
      aState : TReaderState;
    begin
      RequireIdentifier( 'wires');
      RequireLParen;
      Done := False;

      while not Done
      do begin
        SaveState( aState);
        ReadToken;

        if ( FTokenType = ttIdentifier) and SameText( FToken, 'wire')
        then begin
          RestoreState( aState);
          ReadWire;
        end
        else if ( FTokenType = ttRParen)
        then Done := True
        else ExpectedIdentifierOrSingle( 'wire', ')');
      end;
    end;


    procedure   TPatchReader.ReadDetail;
    var
      aType    : string;
      aName    : string;
      aValue   : Integer;
      aString  : string;
      isString : Boolean;
      aState   : TReaderState;
      aLocked  : Boolean;
      aCC      : Byte;
    begin
      isString := False;
      RequireIdentifier( 'detail');
      RequireLParen;
      aLocked := False;
      aCC     := 0;
      aType   := ReadClass;
      aName   := ReadName;
      SaveState( aState);

      if not ReadValue( aValue)
      then begin
        RestoreState( aState);
        isString := ReadStringValue( aString);
        if not isString
        then ExpectedStringOrNumber;
      end;

      ReadLockedOption( aLocked);
      ReadCCOption( aCC);
      ScanRParen;                   // this was RequireRParen, but we allow for as of yet unforseen options to be added and be silently ignored

      if Assigned( FModule)
      then begin

        // First set a possible lock, so a later position setting for a knob (SetKnobValue) will not change a possibly
        // associated display (for a locked knob the display holds the value, if present).

        if aLocked
        then FModule.SetLockOn( aType, aName);

        FModule.SetCCOn( aType, aName, aCC);

        if isString
        then FModule.SetDetailValue( aType, aName, aString)
        else FModule.SetKnobValue  ( aType, aName, aValue );
      end;
    end;


    procedure   TPatchReader.ReadDetails;
    var
      Done   : Boolean;
      aState : TReaderState;
    begin
      RequireIdentifier( 'details');
      RequireLParen;
      Done := False;

      while not Done
      do begin
        SaveState( aState);
        ReadToken;

        if ( FTokenType = ttIdentifier) and SameText( FToken, 'detail')
        then begin
          RestoreState( aState);
          ReadDetail;
        end
        else if ( FTokenType = ttRParen)
        then Done := True
        else ExpectedIdentifierOrSingle( 'detail', ')');
      end;
    end;


    procedure   TPatchReader.ReadModule; // virtual;
    var
      aModuleClass   : string;
      aModuleName    : string;
      aModuleTitle   : string;
      aModuleType    : Integer;
      aModuleLeft    : Integer;
      aModuleTop     : Integer;
      aModuleColor   : TColor;
      HasCustomColor : Boolean;
    begin
      RequireIdentifier( 'module');
      RequireLParen;
      aModuleClass   := ReadClass;
      aModuleName    := ReadName;
      aModuleTitle   := ReadTitle;
      aModuleType    := ReadType;
      aModuleLeft    := ReadLeft;
      aModuleTop     := ReadTop;
      HasCustomColor := ReadColorOption( aModuleColor);
      FModule        := FormStore.CreateModule( FWirePanel, aModuleType, NO_DRAG, False);

      if Assigned( FModule)
      then begin
        try
          with FModule
          do begin
            Name  := aModuleName;
            Title := aModuleTitle;
            Left  := aModuleLeft;
            Top   := aModuleTop;

            if HasCustomColor
            then Color := aModuleColor
            else Color := TColor( $ffffffff);
          end;

          ReadDetails;
          RequireRParen;
        except
          FreeAndNil( FModule);
          raise;
        end;
      end
      else ReadErrorFmt( 'Could not create module "%s" of type %d with title "%s"', [ aModuleName, aModuleType, aModuleTitle]);
    end;


    procedure   TPatchReader.ReadModules; // virtual;
    var
      Done   : Boolean;
      aState : TReaderState;
    begin
      RequireIdentifier( 'modules');
      RequireLParen;
      Done := False;

      while not Done
      do begin
        SaveState( aState);
        ReadToken;

        if ( FTokenType = ttIdentifier) and SameText( FToken, 'module')
        then begin
          RestoreState( aState);
          ReadModule;
        end
        else if ( FTokenType = ttRParen)
        then Done := True
        else ExpectedIdentifierOrSingle( 'module', ')');
      end;
    end;


    procedure   TPatchReader.Read; // virtual;
    var
      aPanelClass : string;
    begin
      FLineNr  := 1;
      FOffset  := 1;
      FPos     := 1;
      RequireIdentifier( 'panel');
      RequireLParen;
      aPanelClass := ReadClass;

      if SameText( aPanelClass, 'TKnobsWirePanel')
      then begin
        FWirePanel.LockUndoRedo;
        try
          FWirePanel.Name    := ReadName;
          FWirePanel.Title   := ReadTitle;
          FWirePanel.Version := ReadVersion;
          FPatchVersion      := FWirePanel.Version;

          if not ( FPatchVersion in [ 2, 3, 4])
          then ReadErrorFmt( 'can''t handle patch version %d', [ FPatchVersion]);

          ReadModules;
          ReadWires;
          RequireRParen;
        finally
          FWirePanel.UnLockUndoRedo;
        end;
      end
      else ReadErrorFmt( 'can''t handle objects of class ''%s'' - ''TKnobsWirePanel'' expected', [ aPanelClass]);
    end;


    function    TPatchReader.TryRead: Boolean; // virtual;
    begin
      Result := False;
      try
        Read;
        Result := True;
      except
        on E: EKnobsCorruptPatch
        do begin
          if Assigned( FOnReadError)
          then FOnReadError( Self, E.Message, Result)
          else raise;
        end;
        on E: Exception
        do KilledException( E);
      end;
    end;


//  public

    constructor TPatchReader.Create; // virtual;
    begin
      inherited;
    end;


    destructor  TPatchReader.Destroy; // override;
    begin
      Clear;
      inherited;
    end;


    function    TPatchReader.GetWarnOnStructMismatch: Boolean;
    begin
      Result := FWarnOnStructMismatch;
    end;


    procedure   TPatchReader.SetWarnOnStructMismatch( aValue: Boolean);
    begin
      FWarnOnStructMismatch := aValue;
    end;


    procedure   TPatchReader.ReadParams(
      const aData      : string;
      const aWirePanel : TKnobsWirePanel;
      ReadStrict       : Boolean
    ); // virtual;
    // This copies to aWirePanel.
    // The internal panel FWirePanel is read from the clipboard, and when the structure of FWirePanel
    // matches the structure of the selected modules in aWirePanel then copy over the parameters
    // from FWirePanel to aWirePanel.
    begin
      if assigned( aWirePanel)
      then begin
        FData := aData;
        SetWirePanel( TKnobsWirePanel.Create( nil));

        if TryRead
        then begin
          aWirePanel.BeginStateChange;

          try
            if not aWirePanel.CopyParamsFrom( FWirePanel, ReadStrict) and WarnOnStructMismatch and ReadStrict
            then Warning( STRUCT_MISMATCH_NO_PASTE);
          finally
            aWirePanel.EndStateChange( False);
          end;
        end;
      end;
    end;


    procedure   TPatchReader.ReadString(
      const aData      : string;
      const aWirePanel : TKnobsWirePanel;
      aDistance        : TPoint;
      ReadStrict       : Boolean;
      MustDrag         : Boolean;
      CopyMidiCC       : Boolean;
      aReadMode        : TKnobsReadMode = rmReplace;
      const aFileName  : string         = ''
    ); // virtual;
    begin
      if Assigned( aWirePanel)
      then begin
        FFileName := aFileName;
        FData     := aData;
        FReadMode := aReadMode;
        SetWirePanel( TKnobsWirePanel.Create( nil));

        if TryRead
        then begin
          aWirePanel.BeginStateChange;

          try
            if FReadMode = rmParams
            then begin
              if not aWirePanel.CopyParamsFrom( FWirePanel, ReadStrict) and WarnOnStructMismatch and ReadStrict
              then Warning( STRUCT_MISMATCH_NO_PASTE);
            end
            else begin
              if FReadMode = rmReplace
              then aWirePanel.FreeModules( nil);

              aWirePanel.AddModules(
                FWirePanel,
                aDistance,
                DO_RESTACK,
                DO_REDRAW_WIRES,
                MustDrag,
                CopyMidiCC
              );

              if aReadMode = rmReplace
              then begin
                aWirePanel.Title   := FWirePanel.Title;
                aWirePanel.Version := FWirePanel.Version;
              end;
            end;
          finally
            aWirePanel.EndStateChange( FReadMode <> rmParams);
          end;
        end;
      end;
    end;


    procedure   TPatchReader.ReadStrings(
      const aData      : TStrings;
      const aWirePanel : TKnobsWirePanel;
      ReadStrict       : Boolean;
      MustDrag         : Boolean;
      CopyMidiCC       : Boolean;
      aReadMode        : TKnobsReadMode = rmReplace;
      const aFileName  : string    = ''
    ); // virtual;
    begin
      ReadString( aData.Text, aWirePanel, Point( 0, 0), True, MustDrag, CopyMidiCC, aReadMode, aFileName);
    end;


    procedure   TPatchReader.ReadFile(
      const aFileName  : string;
      const aWirePanel : TKnobsWirePanel;
      ReadStrict       : Boolean;
      MustDrag         : Boolean;
      CopyMidiCC       : Boolean;
      aReadMode        : TKnobsReadMode = rmReplace
    ); // virtual;
    var
      aData : TStrings;
    begin
      aData := TStringList.Create;

      try
        aData.LoadFromFile( aFileName);
        ReadStrings( aData, aWirePanel, ReadStrict, MustDrag, CopyMidiCC, aReadMode, aFileName);
      finally
        aData.DisposeOf;
      end;
    end;


    procedure   TPatchReader.ImportFile(
      const aFileName  : string;
      const aWirePanel : TKnobsWirePanel;
      ReadStrict       : Boolean;
      MustDrag         : Boolean;
      CopyMidiCC       : Boolean;
      aReadMode        : TKnobsReadMode = rmReplace
    ); // virtual;
    var
      S     : string;
      aData : TStrings;
    begin
      aData := TStringList.Create;

      try
        aData.LoadFromFile( aFileName);
        S := aData.Text;

        if Pos( 'Panel(', S) = 1
        then begin
          try
            ReadString( S, aWirePanel, Point( 0, 0), ReadStrict, MustDrag, CopyMidiCC, aReadMode, '<clip>');
          except
            // Ignore exceptions - prolly something else was being pasted ..
            on E: Exception
            do KilledException( E);
          end;
        end;
      finally
        aData.DisposeOf;
      end;
    end;


    procedure   TPatchReader.ReadFromClipboard(
      const aWirePanel : TKnobsWirePanel;
      ReadStrict       : Boolean;
      MustDrag         : Boolean;
      CopyMidiCC       : Boolean;
      aReadMode        : TKnobsReadMode = rmReplace
    ); // virtual;
    //* Reads a wirepanel from the clipboard and copies the modules etc.
    //* found in it to the aWirePanel parameter. Existing modules
    //* are deleted when aReadMode = rmReplace, otherwise the new modules
    //* are added to the wire panel. The modules pasted should be set into
    //* drag mode, see __2 for other cases (in Knobs2013).
    var
      S : string;
    begin
      S := Clipboard.AsText;

      if Pos( 'Panel(', S) = 1
      then begin
        try
          ReadString( S, aWirePanel, Point( 0, 0), ReadStrict, MustDrag, CopyMidiCC, aReadMode, '<clip>');
        except
          // Ignore exceptions - prolly something else was being pasted ..
          on E: Exception
          do KilledException( E);
        end;
      end;
    end;


    procedure   TPatchReader.ParamsFromFile(
      const aFileName  : string;
      const aWirePanel : TKnobsWirePanel;
      ReadStrict       : Boolean
    ); // virtual;
    //* Reads a wirepanel from a file, and when that panel matches the structure
    //* of the current patch selection in aWirePanel copy the parameter values
    //* from the clipboard over to the selection of aWirePanel.
    begin
      ReadFile( aFileName, aWirePanel, ReadStrict, NO_DRAG, NO_COPY_MIDI, rmParams);
    end;


    procedure   TPatchReader.ParamsFromClipboard(
      const aWirePanel : TKnobsWirePanel;
      ReadStrict       : Boolean
    ); // virtual;
    //* Reads a wirepanel from the clipboard, and when that panel matches the structure
    //* of the current patch selection in aWirePanel copy the parameter values
    //* from the clipboard over to the selection of aWirePanel.
    var
      S : string;
    begin
      S := Clipboard.AsText;

      if Pos( 'Panel(', S) = 1
      then ReadParams( S, aWirePanel, ReadStrict);
    end;


{ ========
  TPatchWriter = class( TInterfacedObject, IKnobsPatchWriter)
  strict private
  var
    FFileName     : string;
    FWriteMode    : TKnobsWriteMode;
    FWirePanel    : TKnobsWirePanel;
    FModule       : TKnobsCustomModule;
    FComponent    : TComponent;
    FWire         : TKnobsWire;
    FOnWriteError : TOnPatchWriteError;
    FResult       : string;
    FLineBreak    : string;
  public
    property    FileName  : string read FFileName;
    property    LineBreak : string read FLineBreak write FLineBreak;
  public
    property    OnWriteError : TOnPatchWriteError read FOnWriteError write FOnWriteError;
  protected
}

    procedure   TPatchWriter.Clear; // virtual;
    begin
      FFileName     := '';
      FWirePanel    := nil;
      FResult       := '';
    end;


    procedure   TPatchWriter.Write( anInd: Integer; const S: string);
    begin
      FResult := FResult + Indent( anInd) + S;
    end;


    procedure   TPatchWriter.WriteLn( anInd: Integer; const S: string);
    begin
      Write( anInd, S);
      Write( 0, LineBreak);
    end;


    procedure   TPatchWriter.WriteWire( anInd: Integer); // virtual;
    begin
      with FWire
      do begin
        if ( FWriteMode = wmAll) or ( Source.Module.Selected and Destination.Module.Selected)
        then WriteLn( anInd, AsString);
      end;
    end;


    procedure   TPatchWriter.WriteWires( anInd: Integer); // virtual;
    var
      i : Integer;
    begin
      with FWirePanel
      do begin
        for i := 0 to WireCount - 1
        do begin
          FWire := Wire[ i];
          WriteWire( anInd);
        end;
      end;
    end;


    procedure   TPatchWriter.WriteDetailData( anInd: Integer; const aValue, aCC: Integer; aLocked: Boolean; const aSuffix: string); //  overload; virtual;
    begin
      if aLocked
      then begin
        WriteLn(
          anInd,
          Format(
            'detail( class = ''%s'' name = ''%s%s'' value = %s locked cc = %d)',
            [
              FComponent.ClassName,
              FComponent.Name,
              aSuffix,
              IntToStr( aValue),
              aCC
            ]
          )
        );
      end
      else begin
        WriteLn(
          anInd,
          Format(
            'detail( class = ''%s'' name = ''%s%s'' value = %s cc = %d)',
            [
              FComponent.ClassName,
              FComponent.Name,
              aSuffix,
              IntToStr( aValue),
              aCC
            ]
          )
        );
      end;
    end;


    procedure   TPatchWriter.WriteDetailData( anInd: Integer; const aValue: string); // overload; virtual;
    begin
      WriteLn(
        anInd,
        Format(
          'detail( class = ''%s'' name = ''%s'' value = ''%s'')',
          [
            FComponent.ClassName,
            FComponent.Name,
            QuoteEscape( aValue)
          ]
        )
      );
    end;


    procedure   TPatchWriter.WriteDetail( anInd: Integer); // virtual;
    var
      aDisplay : TKnobsDisplay;
    begin
      if FComponent is TKnobsValuedControl
      then begin
        if TKnobsValuedControl( FComponent).Locked
        then begin
          // For a locked knob write both the knob position and the associated
          // display's caption.
          WriteDetailData( anInd, TKnobsValuedControl( FComponent).KnobPosition, TKnobsValuedControl( FComponent).AssignedMIDICC, True, '');

          if ( FComponent is TKnobsKnob)
          then begin
            aDisplay := TKnobsKnob( FComponent).Display;

            if Assigned( aDisplay)
            then begin
              FComponent := aDisplay;
              WriteDetailData( anInd, aDisplay.Caption);
            end;
          end;
        end
        else WriteDetailData( anInd, TKnobsValuedControl( FComponent).KnobPosition, TKnobsValuedControl( FComponent).AssignedMIDICC, False, '');
      end
      else if ( FComponent is TKnobsDisplay) and ( TKnobsDisplay( FComponent).MustSave)
      then WriteDetailData( anInd, TKnobsDisplay( FComponent).Caption)
      else if ( FComponent is TKnobsFileSelector) and ( TKnobsFileSelector( FComponent).MustSave)
      then WriteDetailData( anInd, TKnobsFileSelector( FComponent).DataFileName)
      else if ( FComponent is TKnobsXYControl)
      then begin
        WriteDetailData( anInd, TKnobsXYControl( FComponent).KnobPositionX, 0, False, 'x');
        WriteDetailData( anInd, TKnobsXYControl( FComponent).KnobPositionY, 0, False, 'y');
      end
      else if ( FComponent is TKnobsConnector) and ( TKnobsConnector( FComponent).WireColor <> TKnobsConnector( FComponent).DefaultWireColor)
      then WriteDetailData( anInd, IntToStr( Cardinal( TKnobsConnector( FComponent).WireColor))); // Write Color as a string value, that's easier to parse on patch reads
    end;


    procedure   TPatchWriter.WriteDetails( anInd: Integer); // virtual;
    var
      i : Integer;
    begin
      with FModule
      do begin
        for i := 0 to ComponentCount - 1
        do begin
          FComponent := Components[ i];
          WriteDetail( anInd);
        end;
      end;
    end;


    procedure   TPatchWriter.WriteModule( anInd, anXOffset, anYOffset: Integer); // virtual;
    begin
      with FModule
      do begin
        if ( FWriteMode = wmAll) or Selected
        then begin
          WriteLn( anInd + 0,         'module('                               );
          WriteLn( anInd + 1, Format(   'class = ''%s''', [ ClassName       ]));
          WriteLn( anInd + 1, Format(   'name  = ''%s''', [ Name            ]));
          WriteLn( anInd + 1, Format(   'title = ''%s''', [ Title           ]));
          WriteLn( anInd + 1, Format(   'type  = %d'    , [ ModuleType      ]));
          WriteLn( anInd + 1, Format(   'left  = %d'    , [ Left + anXOffset]));
          WriteLn( anInd + 1, Format(   'top   = %d'    , [ Top  + anYOffset]));
          WriteLn( anInd + 1, Format(   'color = %d'    , [ Cardinal( Color)]));
          WriteLn( anInd + 1,           'details('                            );

          WriteDetails( anInd + 2);

          WriteLn( anInd + 1,           ')'                                   );
          WriteLn( anInd + 0,         ')'                                     );
        end;
      end;
    end;


    procedure   TPatchWriter.WriteModules( anInd: Integer); // virtual;
    var
      i         : Integer;
      anXOffset : Integer;
      anYOffset : Integer;
    begin
      with FWirePanel
      do begin
        // Find largest Left and Top offsets needed to avoid writing a patch with negative Left or Top indices.
        // On write all the modules will be shifted by these offsets.
        anXOffset := 0;
        anYOffset := 0;

        for i := 0 to ModuleCount - 1
        do begin
          FModule := Module[ i];
          if - FModule.Left > anXOffset
          then anXOffset := - FModule.Left;
          if - FModule.Top > anYOffset
          then anYOffset := - FModule.Top;
        end;

        // Write the moddules shifted by the calculated offsets.
        for i := 0 to ModuleCount - 1
        do begin
          FModule := Module[ i];
          WriteModule( anInd, anXOffset, anYOffset);
        end;
      end;
    end;


    procedure   TPatchWriter.WritePatch; // virtual;
    var
      anInd : Integer;
    begin
      with FWirePanel
      do begin
        anInd := 0;
        WriteLn( anInd + 0,         'Panel('                                   );
        WriteLn( anInd + 1, Format(   'class   = ''%s''', [ ClassName        ]));
        WriteLn( anInd + 1, Format(   'name    = ''%s''', [ Name             ]));
        WriteLn( anInd + 1, Format(   'title   = ''%s''', [ Title            ]));
        WriteLn( anInd + 1, Format(   'version = %d'    , [ KnobsPatchVersion]));
        WriteLn( anInd + 1,           'modules('                               );

        WriteModules( anInd + 2);

        WriteLn( anInd + 1,           ')'                                      );
        WriteLn( anInd + 1,           'wires('                                 );

        WriteWires( anInd + 2);

        WriteLn( anInd + 1,           ')'                                      );
        WriteLn( anInd + 0,         ')'                                        );
      end;
    end;


    function    TPatchWriter.TryWrite: Boolean; // virtual;
    begin
      Result := False;

      try
        WritePatch;
        Result := True;
      except on E: Exception
      do
        if Assigned( FOnWriteError)
        then FOnWriteError( Self, E.Message, Result)
        else raise;
      end;
    end;


//  public

    constructor TPatchWriter.Create; // virtual;
    begin
      inherited;

      LineBreak := #13#10;
    end;


    destructor  TPatchWriter.Destroy; // override;
    begin
      Clear;

      inherited;
    end;


    function    TPatchWriter.WriteString(
      const aWirePanel: TKnobsWirePanel;
      const aFileName : string     = '';
      aWriteMode      : TKnobsWriteMode = wmAll
    ): string; // virtual;
    begin
      FResult    := '';
      FFileName  := aFileName;
      FWirePanel := aWirePanel;
      FWriteMode := aWriteMode;

      if TryWrite
      then Result := FResult;
    end;


    function    TPatchWriter.WriteStrings(
      const aWirePanel : TKnobsWirePanel;
      const aFileName  : string     = '';
      aWriteMode       : TKnobsWriteMode = wmAll
    ): TStrings; // virtual;
    begin
      Result := TStringList.Create;

      try
        Result.Text := WriteString( aWirePanel, aFileName, aWriteMode);
      except
        Result.DisposeOf;
        raise;
      end;
    end;


    procedure   TPatchWriter.WriteFile(
      const aWirePanel : TKnobsWirePanel;
      const aFileName  : string;
      aWriteMode       : TKnobsWriteMode = wmAll
    ); // virtual;
    var
      aStrings   : TStrings;
      anOldTitle : string;
    begin
      if Assigned( aWirePanel)
      then begin
        anOldTitle := aWirePanel.Title;

        if aWriteMode = wmParams
        then aWirePanel.Title := 'params';

        try
          aStrings := WriteStrings( aWirePanel, aFileName, aWriteMode);
        finally
          aWirePanel.Title := anOldTitle;
        end;

        try
          aStrings.SaveToFile( aFileName);
        finally
          aStrings.DisposeOf;
        end;
      end;
    end;


    procedure   TPatchWriter.WriteToClipboard(
      const aWirePanel : TKnobswirePanel;
      aWriteMode       : TKnobsWriteMode = wmAll
    ); // virtual;
    begin
      ClipBoard.AsText := WriteString( aWirePanel, '', aWriteMode);
    end;


end.
